/******************************************************************************
*
* project name:    TI-Mahjongg Solitaire
* file name:       timah.c
* initial date:    05/11/2000
* authors:         thomas.nussbaumer@gmx.net
*                  marcos.lopez@wol.es
*                  lionel_debroux@yahoo.fr
*
* description:     main program
*
* NOTE:  No animals or trees were harmed for the stones of this Mahjongg game
*
*
* $Id: timah.c,v 1.2 2004/07/15 13:02:39 DEBROUX Lionel Exp $
*
******************************************************************************/

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#include <tigcclib.h>
#include "timah.h"


#include "board.c"
#include "cursor.c"
#include "clock.c"
#include "logo.c"

/*---------------------------------------------------------------------------*/
/* some globals                                                              */
/*---------------------------------------------------------------------------*/
SCR_RECT fullscreen = {{0,0,239,127}}; // helper variable for clipping


/*---------------------------------------------------------------------------*/
/* some defines for drawing functions                                        */
/*---------------------------------------------------------------------------*/
//#define COLOR_WHITE      0
//#define COLOR_LIGHTGRAY  1
//#define COLOR_DARKGRAY   2
//#define COLOR_BLACK      3
//#define LIGHT_PLANE      0
//#define DARK_PLANE       1
//#define RECT_EMPTY       0
//#define RECT_FILLED      1
//#define A_CENTERED       0x40
//#define A_SHADOWED       0x80

int cursor_x = 0;
int cursor_y = 0;


/*===========================================================================*/
/* utility function to draw a string to both planes                          */
/* NOTE: there exists two new styles which are handled as flags              */
/*       A_CENTERED ... the given font will be used and the string will be   */
/*                      drawn centered on the screen                         */
/*       A_SHADOWED ... if this flag is set in the attributes a lightgray    */
/*                      shadow will be drawn with +1 pixel offset in all     */
/*                      two directions                                       */
/*===========================================================================*/
void GrayDrawStrExt_R(short x,short y,const char* s,short attr,short font) {
    short do_shadow = attr & A_SHADOWED;
    attr &= ~A_SHADOWED;

    short oldfont = FontGetSys();
    FontSetSys(font);

    if (attr & A_CENTERED) {
        x = (LCD_WIDTH-DrawStrWidth(s,font))/2;
        attr &= ~A_CENTERED;
    }

    GraySetAMSPlane(DARK_PLANE);
    DrawStr(x,y,s,attr);
    GraySetAMSPlane(LIGHT_PLANE);
    DrawStr(x,y,s,attr);

    if (do_shadow) DrawStr(x+1,y+1,s,A_NORMAL);
    FontSetSys(oldfont);
}


/*===========================================================================*/
/* draw a line with given color from (x0,y0) to (x1,y1)                      */
/*===========================================================================*/
#define DrawColorLine GrayDrawLine

/*===========================================================================*/
/* draws a rectangle with given color                                        */
/* if fill == 1 the rectangle is filled                                      */
/*===========================================================================*/
#define DrawColorRect GrayDrawRect


/*===========================================================================*/
/* draws a popup window                                                      */
/*===========================================================================*/
void DrawPopup(int x0,int y0,int x1, int y1) {
    DrawColorRect(x0,y0,x1,y1,COLOR_BLACK,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_DARKGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_LIGHTGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_WHITE,RECT_FILLED);
}


/*===========================================================================*/
/* outputs the time to the statusline                                        */
/*===========================================================================*/
void PrintTime(void) {
    char buffer[16];
    sprintf(buffer,"%02d:%02d:%02d", hours, minutes, seconds);
    GrayDrawStrExt_R(130,95,buffer,A_REPLACE,F_4x6);
}


/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*                                                                           */
/* additionally this function handles the CALCULATOR OFF combinations        */
/*===========================================================================*/
unsigned short GetUserInput(void) {
    void           *kbq = kbd_queue();
    unsigned short  key;
    short           gray = GrayCheckRunning();

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (!gray) idle();
            else pokeIO(0x600005,0b10111);
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
            if (outputtime) {
                outputtime = 0;
                PrintTime();
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires
        // [2nd] + [ON] & [diamond]+[ON]
        // calculator off
        //---------------------------------------

        if (key == KEY_OFF || key == KEY_OFF2) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & ~(unsigned short)0x800; // map out key repeat flag
        }
    }
}

/*===========================================================================*/
/* treatment of cursor position changement                                   */
/*===========================================================================*/
void CursorUpdate(void) {
    int toplevel;
    int pos;

    if (cursor_x >= BOARD_SIZE_X) cursor_x = 0;
    if (cursor_x < 0)             cursor_x = BOARD_SIZE_X-1;
    if (cursor_y >= BOARD_SIZE_Y) cursor_y = 0;
    if (cursor_y < 0)             cursor_y = BOARD_SIZE_Y-1;

    toplevel = GetTopLayer(cursor_x,cursor_y);
    if (toplevel == -1) toplevel = 0;

    RestoreOldCursorPos();
    BackupCursorPos(toplevel,cursor_x,cursor_y);
    OutputCursor(toplevel,cursor_x,cursor_y);

    if (sel_layer == -1) pos = 0;
    else                 pos = 1;

    UpdateHelpField(GetStone(toplevel,cursor_x,cursor_y),pos);

    {
        char buf[100];
        sprintf(buf,"[F5 for Help]      Pairs Left = %d",pairs_left);

        GrayDrawStrExt_R(0,95,buf,A_REPLACE,F_4x6);
    }
}


unsigned char* helppage[] = {
    "To win this Game You have to remove all",
    "Stones from the Board.",
    "Stones can only be removed in Pairs and if the",
    "following Conditions hold true:",
    "",
    "(1) Both Stones have the same Graphics",
    "(2) Each Stone can be moved left OR right at",
    "    least one Field",
    "",
    "Use the Cursor Keys and [ENTER] or [F1] to",
    "select the Stones.",
    "",
    "[CLEAR] to Reset  [+][-] to change Setup",
    NULL};

/*===========================================================================*/
/* displays help page                                                        */
/*===========================================================================*/
void DisplayHelp() {
    int i = 0;

    running    = 0;
    outputtime = 0;

    memset(GetPlane(0),0,LCD_SIZE);
    memset(GetPlane(1),0,LCD_SIZE);


    GrayDrawStrExt_R(0,0,TITLE_AND_VERSION" by TICT",A_REPLACE | A_SHADOWED,F_6x8);

    while (helppage[i]) {
        GrayDrawStrExt_R(0,10+i*7,helppage[i],A_REPLACE,F_4x6);
        i++;
    }
    GetUserInput();

    running    = 1;
    outputtime = 1;
}


/*===========================================================================*/
/* initialize a new game (returns 1 if okay)                                 */
/*===========================================================================*/
int InitGame(void) {
    int i=20;
    randomize();

    resettime = 1;
    running   = 1;
    cursor_x  = 0;
    cursor_y  = 0;

    //------------------------------------------------------------
    // for complicated setups we'll try at least 20 times to setup
    //------------------------------------------------------------
    do {
        if (!InitBoard()) return 0;
    } while (!CheckPossibleMovesLeft() && i--);

    if (i<=0) return 0;

    DrawBoard();
    PrintTime();
    CursorUpdate();
    return 1;
}


/*===========================================================================*/
/* draws a message                                                           */
/*===========================================================================*/
void DrawMessage(char* msg) {
    int sy = LCD_HEIGHT/2 - 2 + 44;

    DrawPopup(1,sy-6,LCD_WIDTH-1,sy+7);
    GrayDrawStrExt_R(0,sy-3-offsety,msg,A_REPLACE | A_CENTERED | A_SHADOWED,F_6x8);

    GetUserInput();
}


/*===========================================================================*/
/* handles and heavy error                                                   */
/*===========================================================================*/
void ErrorTreatment(void) {
    DrawMessage("ERROR: cannot initialize");
    FreeVariables();
    SetIntVec(AUTO_INT_5, oldint5);
    GrayOff();
}


/*===========================================================================*/
/* as usual our main routine ...                                             */
/*===========================================================================*/
void _main(void) {
    int key;

    if (!GrayOn()) return;

    if (!AllocateVariables()) {
        GrayOff();
        FreeVariables();
        return;
    }

    outputtime = mseconds50 = seconds = minutes = hours = 0;

    Logo();

    if (!InitGame()) {
        DrawMessage("ERROR: cannot initialize");
        GrayOff();
        FreeVariables();
        return;
    }

    oldint5 = GetIntVec(AUTO_INT_5);
    SetIntVec(AUTO_INT_5, myint5handler);

    do {
        key = GetUserInput();

        if (key == KEY_UP) {
            cursor_y--;
            CursorUpdate();
        }
        else if (key == KEY_DOWN) {
            cursor_y++;
            CursorUpdate();
        }
        else if (key == KEY_LEFT) {
            cursor_x--;
            CursorUpdate();
        }
        else if (key == KEY_RIGHT) {
            cursor_x++;
            CursorUpdate();
        }
        else if (key == KEY_ENTER || key == KEY_F1) {
            if (HandleSelect(cursor_x,cursor_y)) {
                //---------------------------
                // !!! THE PLAYER HAS WON !!!
                //---------------------------
                char msg[100];
                running    = 0;
                outputtime = 0;
                sprintf(msg,"You've won in %02d:%02d:%02d !", hours, minutes, seconds);
                DrawMessage(msg);
                if (!InitGame()) {
                    ErrorTreatment();
                    return;
                }
            }
            else if (!CheckPossibleMovesLeft()) {
                //----------------------------
                // !!! THE PLAYER HAS LOST !!!
                //----------------------------
                char msg[100];
                running    = 0;
                outputtime = 0;
                sprintf(msg,"You lost (%d Pairs left)!",pairs_left);
                DrawMessage(msg);
                if (!InitGame()) {
                    ErrorTreatment();
                    return;
                }
            }
            else {
                CursorUpdate();
            }
        }
        else if (key == KEY_CLEAR || key == '+' || key == '-') {
            //------------------------
            // handle setup change ...
            //------------------------
            if (key == '+')      used_setup++;
            else if (key == '-') used_setup--;
            if (used_setup < 0)          used_setup = NR_SETUPS-1;
            if (used_setup >= NR_SETUPS) used_setup = 0;

            if (!InitGame()) {
                ErrorTreatment();
                return;
            }
        }
        else if (key == KEY_F5) {
            DisplayHelp();
            DrawBoard();
            CursorUpdate();
        }
    }
    while (key != KEY_ESC);

    SetIntVec(AUTO_INT_5, oldint5);
    FreeVariables();
    GrayOff();
}


//=============================================================================
// Revision History
//=============================================================================
//
// $Log: timah.c,v $
// Revision 1.2  2004/07/15 13:02:39  DEBROUX Lionel
// generic commit
//
// Revision 1.1  2000/11/14 19:14:01  Thomas Nussbaumer
// initial version
//
//
